#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <vmath.h>

#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

struct ball {
	vec2_t pos;
	vec2_t vel;
	double mass;
	int fixed;
};

void redraw(void);
void timer(int val);
vec2_t spring_force(vec2_t vec, double slen, double sk);
void reshape(int x, int y);
void keyb(unsigned char key, int x, int y);
void button(int bn, int state, int x, int y);
void motion(int x, int y);

int xsz, ysz;
int upd_interval = 33;

int pt_count = 7;
struct ball *rope;
double spring_len = 0.075;
double spring_k = 0.8;
double ball_mass = 0.925;
double grav = 0.01;
double friction = 0.8;

int stop_eval;

int main(int argc, char **argv)
{
	int i;

	glutInitWindowSize(800, 600);
	glutInit(&argc, argv);

	for(i=1; i<argc; i++) {
		if(argv[i][0] == '-' && argv[i][2] == 0) {
			switch(argv[i][1]) {
			case 'n':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-n must be followed by the number of balls\n");
					return 1;
				}
				pt_count = atoi(argv[i]);
				break;

			case 'k':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-k must be followed by the spring constant\n");
					return 1;
				}
				spring_k = atof(argv[i]);
				break;

			case 'm':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-m must be followed by the mass of the balls\n");
					return 1;
				}
				ball_mass = atof(argv[i]);
				break;

			case 'g':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-g must be followed by the gravity coefficient\n");
					return 1;
				}
				grav = atof(argv[i]);
				break;

			case 'l':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-l must be followed by the rope segment length\n");
					return 1;
				}
				spring_len = atof(argv[i]);
				break;

			case 'f':
				if(!isdigit(argv[++i][0])) {
					fprintf(stderr, "-f must be followed by the friction coefficient\n");
					return 1;
				}
				friction = atof(argv[i]);
				break;

			default:
				fprintf(stderr, "unrecognized option: %s\n", argv[i]);
				return 1;
			}
		} else {
			fprintf(stderr, "unexpected argument: %s\n", argv[i]);
			return 1;
		}
	}

	glutInitDisplayMode(GLUT_RGB | GLUT_DEPTH | GLUT_DOUBLE);
	glutCreateWindow("foo");

	glutDisplayFunc(redraw);
	glutReshapeFunc(reshape);
	glutKeyboardFunc(keyb);
	glutMotionFunc(motion);
	glutMouseFunc(button);

	if(!(rope = malloc(pt_count * sizeof *rope))) {
		perror("malloc failed");
		return 1;
	}
	memset(rope, 0, pt_count * sizeof *rope);
	for(i=0; i<pt_count; i++) {
		rope[i].pos.x = 0;
		rope[i].pos.y = -(i * spring_len);
		rope[i].mass = 1.0;
		rope[i].fixed = i > 0 ? 0 : 1;
	}

	glClearColor(0, 0, 0, 0);
	glClearDepth(1.0);

	glEnable(GL_POINT_SMOOTH);

	glutMainLoop();
	return 0;
}


void redraw(void)
{
	int i;
	unsigned int frm_time, frm_start = glutGet(GLUT_ELAPSED_TIME);

	glClear(GL_COLOR_BUFFER_BIT);

	glLineWidth(1.0);
	glBegin(GL_LINE_STRIP);
	glColor3f(0.0, 1.0, 0.0);
	for(i=0; i<pt_count; i++) {
		glVertex2f(rope[i].pos.x, rope[i].pos.y);
	}
	glEnd();

	glPointSize(4.0);
	glBegin(GL_POINTS);
	glColor3f(1.0, 0.0, 0.0);
	for(i=0; i<pt_count; i++) {
		glVertex2f(rope[i].pos.x, rope[i].pos.y);
	}
	glEnd();

	glutSwapBuffers();

	frm_time = glutGet(GLUT_ELAPSED_TIME) - frm_start;
	glutTimerFunc(upd_interval - frm_time, timer, 0);
}

void timer(int val)
{
	int i;

	if(!val && stop_eval) {
		return;
	}

	for(i=0; i<pt_count; i++) {
		vec2_t sf1, sf2;
		vec2_t acc1 = {0, 0}, acc2 = {0, 0};
		vec2_t vel;

		if(rope[i].fixed) {
			continue;
		}
		
		if(i > 0) {
			sf1 = spring_force(v2_sub(rope[i - 1].pos, rope[i].pos), spring_len, spring_k);
			acc1 = v2_scale(sf1, 1.0 / ball_mass);
		}
		
		if(pt_count - i > 1) {
			sf2 = spring_force(v2_sub(rope[i + 1].pos, rope[i].pos), spring_len, spring_k);
			acc2 = v2_scale(sf2, 1.0 / ball_mass);
		}

		vel = v2_add(acc1, acc2);
		vel.y -= grav;

		rope[i].vel = v2_add(v2_scale(rope[i].vel, friction), vel);
	}

	for(i=1; i<pt_count; i++) {
		rope[i].pos = v2_add(rope[i].pos, rope[i].vel);
	}

	glutPostRedisplay();
}

vec2_t spring_force(vec2_t vec, double slen, double sk)
{
	double len = v2_length(vec);
	double off = len - slen;
	double flen = sk * off;

	return v2_scale(vec, flen / len);
}

void reshape(int x, int y)
{
	glViewport(0, 0, x, y);
	xsz = x;
	ysz = y;
}

void keyb(unsigned char key, int x, int y)
{
	switch(key) {
	case 27:
		exit(0);

	case ' ':
		stop_eval = !stop_eval;
		break;

	case 's':
		timer(1);
		break;
	}
}


static struct ball *rp;

void button(int bn, int state, int x, int y)
{
	int i;
	vec2_t curs;

	curs.x = 2.0 * (double)x / (double)xsz - 1.0;
	curs.y = 1.0 - 2.0 * (double)y / (double)ysz;

	if(state == GLUT_DOWN) {
		for(i=0; i<pt_count; i++) {
			if(v2_length_sq(v2_sub(curs, rope[i].pos)) < 0.0005) {
				rp = rope + i;
				rp->fixed = 1;
				rp->vel.x = rp->vel.y = 0;
				return;
			}
		}
	} else {
		if(rp && rp != rope) {
			rp->fixed = 0;
			rp->vel.x = rp->vel.y = 0;
			rp->pos.x = curs.x;
			rp->pos.y = curs.y;
		}
		rp = 0;
	}
}

void motion(int x, int y)
{
	if(rp) {
		rp->pos.x = 2.0 * (double)x / (double)xsz - 1.0;
		rp->pos.y = 1.0 - 2.0 * (double)y / (double)ysz;
	}
}
